# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import fixtures
import logging
import textwrap
import testtools
import voluptuous as vs
from collections import defaultdict
from configparser import ConfigParser

from zuul import model
from zuul.lib.ansible import AnsibleManager
from zuul.configloader import (
    AuthorizationRuleParser, ConfigLoader, safe_load_yaml
)
from zuul.model import Abide, MergeRequest, SourceContext
from zuul.zk.locks import tenant_read_lock

from tests.base import (
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
)
from tests.unit.test_launcher import LauncherBaseTestCase


class TestConfigLoader(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_update_system_config(self):
        """Test if the system config can be updated without a scheduler."""
        sched = self.scheds.first.sched

        # Get the current system config before instantiating a ConfigLoader.
        unparsed_abide, zuul_globals = sched.system_config_cache.get()
        ansible_manager = AnsibleManager(
            default_version=zuul_globals.default_ansible_version)
        loader = ConfigLoader(
            sched.connections, sched.system, self.zk_client, zuul_globals,
            sched.unparsed_config_cache, sched.statsd,
            keystorage=sched.keystore)
        abide = Abide()
        loader.loadTPCs(abide, unparsed_abide)
        loader.loadAuthzRules(abide, unparsed_abide)

        for tenant_name in unparsed_abide.tenants:
            tlock = tenant_read_lock(self.zk_client, tenant_name, self.log)
            # Consider all caches valid (min. ltime -1)
            min_ltimes = defaultdict(lambda: defaultdict(lambda: -1))
            with tlock:
                tenant = loader.loadTenant(
                    abide, tenant_name, ansible_manager, unparsed_abide,
                    min_ltimes=min_ltimes)

                self.assertEqual(tenant.name, tenant_name)

    @okay_tracebacks(' _processCatJob')
    def test_merge_job_cleanup(self):
        # When issuing a full reconfiguration or a tenant validation, we
        # cannot hold the merge jobs in queue as the configloader is
        # explicitly waiting for the cat jobs to complete. Thus, instead
        # of holding the merge jobs we directly fail those cat jobs to
        # trigger the merge request cleanup.
        self.merger_api.failJobs(MergeRequest.CAT)
        self.executor_server.merger_api.failJobs(MergeRequest.CAT)
        self.assertRaises(
            Exception,
            lambda: self.scheds.execute(lambda app: app.sched.validateTenants(
                app.config, {'tenant-one'})),
        )

        self.waitUntilSettled()

        # Even if the merge jobs are canceled, there could be a race
        # condition where an executor picks up a merge request and
        # re-creates the result znode in ZK after the request got
        # deleted. To prevent assertCleanZooKeeper to fail this test
        # even if the cancelation was successful, we have to manually
        # clean up lost merge requests.
        self.merger_api.cleanup()

        # As we are explicitly failing cat jobs we are expecting
        # validateTenants() to fail with an exception.
        # As this catches other exceptions that might be raised when
        # canceling the merge jobs, there is no proper way to validate
        # this apart from checking the test/exception logs for specific
        # messages.
        for record in self._exception_logs:
            if "Unable to cancel job" in record.msg:
                self.fail(
                    f"Exception while canceling merge jobs: {record.msg}")


class TenantParserTestCase(ZuulTestCase):
    create_project_keys = True

    JOB_SET = set([
        'pipeline', 'job', 'semaphore', 'project',
        'project-template', 'nodeset', 'secret', 'queue',
    ])
    PROVIDER_SET = set([
        'image', 'flavor', 'label', 'section', 'provider',
    ])
    UNTRUSTED_SET = JOB_SET
    CONFIG_SET = UNTRUSTED_SET | PROVIDER_SET

    def setupAllProjectKeys(self, config: ConfigParser):
        for project in ['common-config', 'org/project1', 'org/project2']:
            self.setupProjectKeys('gerrit', project)


class TestTenantSimple(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/simple.yaml'

    def test_tenant_simple(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])

        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET, tpc.load_classes)
        self.assertTrue('common-config-job' in tenant.layout.jobs)
        self.assertTrue('project1-job' in tenant.layout.jobs)
        self.assertTrue('project2-job' in tenant.layout.jobs)
        project1_config = tenant.layout.project_configs.get(
            'review.example.com/org/project1')
        self.assertTrue('common-config-job' in
                        project1_config[0].pipelines['check'].job_list.jobs)
        self.assertTrue('project1-job' in
                        project1_config[1].pipelines['check'].job_list.jobs)
        project2_config = tenant.layout.project_configs.get(
            'review.example.com/org/project2')
        self.assertTrue('common-config-job' in
                        project2_config[0].pipelines['check'].job_list.jobs)
        self.assertTrue('project2-job' in
                        project2_config[1].pipelines['check'].job_list.jobs)

    def test_cache(self):
        # A full reconfiguration should issue cat jobs for all repos
        with self.assertLogs('zuul.TenantParser', level='DEBUG') as full_logs:
            self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
            self.waitUntilSettled()
            self.log.debug("Full reconfigure logs:")
            for x in full_logs.output:
                self.log.debug(x)
            self.assertRegexInList(
                r'Submitting cat job (.*?) for gerrit common-config master',
                full_logs.output)
            self.assertRegexInList(
                r'Submitting cat job (.*?) for gerrit org/project1 master',
                full_logs.output)
            self.assertRegexInList(
                r'Submitting cat job (.*?) for gerrit org/project2 master',
                full_logs.output)
            self.assertRegexNotInList(
                r'Using files from cache',
                full_logs.output)

        first = self.scheds.first
        second = self.createScheduler()
        second.start()
        self.assertEqual(len(self.scheds), 2)
        for _ in iterate_timeout(10, "until priming is complete"):
            state_one = first.sched.local_layout_state.get("tenant-one")
            if state_one:
                break

        for _ in iterate_timeout(
                10, "all schedulers to have the same layout state"):
            if (second.sched.local_layout_state.get(
                    "tenant-one") == state_one):
                break

        self.log.debug("Freeze scheduler-1")
        # Start the log context manager for the update test below now,
        # so that it's already in place when we release the second
        # scheduler lock.
        with self.assertLogs('zuul.TenantParser', level='DEBUG'
                             ) as update_logs:
            lock1 = second.sched.layout_update_lock
            lock2 = second.sched.run_handler_lock
            with lock1, lock2:
                # A tenant reconfiguration should use the cache except for the
                # updated project.
                file_dict = {'zuul.d/test.yaml': ''}

                # Now start a second log context manager just for the
                # tenant reconfig test
                with self.assertLogs('zuul.TenantParser', level='DEBUG') \
                     as tenant_logs:
                    A = self.fake_gerrit.addFakeChange(
                        'org/project1', 'master', 'A',
                        files=file_dict)
                    A.setMerged()
                    self.fake_gerrit.addEvent(A.getChangeMergedEvent())
                    self.waitUntilSettled(matcher=[first])
                    self.log.debug("Tenant reconfigure logs:")
                    for x in tenant_logs.output:
                        self.log.debug(x)

                    self.assertRegexNotInList(
                        r'Submitting cat job (.*?) for '
                        r'gerrit common-config master',
                        tenant_logs.output)
                    self.assertRegexInList(
                        r'Submitting cat job (.*?) for '
                        r'gerrit org/project1 master',
                        tenant_logs.output)
                    self.assertRegexNotInList(
                        r'Submitting cat job (.*?) for '
                        r'gerrit org/project2 master',
                        tenant_logs.output)
                    self.assertRegexNotInList(
                        r'Using files from cache',
                        tenant_logs.output)

            # A layout update should use the unparsed config cache
            # except for what needs to be refreshed from the files
            # cache in ZK.
            self.log.debug("Thaw scheduler-1")
            self.waitUntilSettled()
            self.log.debug("Layout update logs:")
            for x in update_logs.output:
                self.log.debug(x)

            self.assertRegexNotInList(
                r'Submitting cat job',
                update_logs.output)
            self.assertRegexNotInList(
                r'Using files from cache for project '
                r'review.example.com/common-config @master.*',
                update_logs.output)
            self.assertRegexInList(
                r'Using files from cache for project '
                r'review.example.com/org/project1 @master.*',
                update_logs.output)
            self.assertRegexNotInList(
                r'Using files from cache for project '
                r'review.example.com/org/project2 @master.*',
                update_logs.output)

    @okay_tracebacks('_cacheTenantYAMLBranch')
    def test_cache_new_branch(self):
        # This tests scheduler startup right after a new branch is
        # created.
        first = self.scheds.first
        lock1 = first.sched.layout_update_lock
        lock2 = first.sched.run_handler_lock
        with lock1, lock2:
            self.create_branch('org/project1', 'stable')
            self.fake_gerrit.addEvent(
                self.fake_gerrit.getFakeBranchCreatedEvent(
                    'org/project1', 'stable'))

            # Start another scheduler and prime it while the existing
            # scheduler is paused and has not processed the new branch
            # creation event.  The new scheduler will encounter a
            # recoverable error with the branch min_ltimes.
            second = self.createScheduler()
            second.start()
            self.assertEqual(len(self.scheds), 2)
            for _ in iterate_timeout(10, "until priming is complete"):
                state_one = first.sched.local_layout_state.get("tenant-one")
                if state_one:
                    break

            for _ in iterate_timeout(
                    10, "all schedulers to have the same layout state"):
                if (second.sched.local_layout_state.get(
                        "tenant-one") == state_one):
                    break
        self.waitUntilSettled()

    def test_variant_description(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        job = tenant.layout.jobs.get("project2-job")
        self.assertEqual(job[0].variant_description, "")
        self.assertEqual(job[1].variant_description, "stable")

    def test_merge_anchor(self):
        to_parse = textwrap.dedent(
            """
            - job:
                name: job1
                vars: &docker_vars
                  registry: 'registry.example.org'

            - job:
                name: job2
                vars:
                  <<: &buildenv_vars
                    image_name: foo
                  <<: *docker_vars

            - job:
                name: job3
                vars:
                  <<: *buildenv_vars
                  <<: *docker_vars
            """)
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        project = list(tenant.config_projects)[0]
        source_context = SourceContext(
            project.canonical_name, project.name, project.connection_name,
            'master', 'zuul.yaml')

        data = safe_load_yaml(to_parse, source_context)
        self.assertEqual(len(data), 3)
        job_vars = [i['job']['vars'] for i in data]
        # Test that merging worked
        self.assertEqual(job_vars, [
            {'registry': 'registry.example.org'},
            {'registry': 'registry.example.org', 'image_name': 'foo'},
            {'registry': 'registry.example.org', 'image_name': 'foo'},
        ])

    def test_deny_localhost_nodeset(self):
        in_repo_conf = textwrap.dedent(
            """
            - nodeset:
                name: localhost
                nodes:
                  - name: localhost
                    label: ubuntu
            """)
        file_dict = {'zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # No job should have run due to the change introducing a config error
        self.assertHistory([])
        self.assertTrue(A.reported)
        self.assertTrue("Nodes named 'localhost' are not allowed."
                        in A.messages[0])

        in_repo_conf = textwrap.dedent(
            """
            - nodeset:
                name: localhost-group
                nodes:
                  - name: ubuntu
                    label: ubuntu
                groups:
                  - name: localhost
                    nodes: ubuntu
            """)
        file_dict = {'zuul.yaml': in_repo_conf}
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # No job should have run due to the change introducing a config error
        self.assertHistory([])
        self.assertTrue(B.reported)
        self.assertTrue("Groups named 'localhost' are not allowed."
                        in B.messages[0])


class TestTenantOverride(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/override.yaml'

    def test_tenant_override(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2', 'org/project4'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET - set(['project']),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(set(['job']), tpc.load_classes)
        self.assertTrue('common-config-job' in tenant.layout.jobs)
        self.assertTrue('project1-job' in tenant.layout.jobs)
        self.assertTrue('project2-job' in tenant.layout.jobs)
        project1_config = tenant.layout.project_configs.get(
            'review.example.com/org/project1')
        self.assertTrue('common-config-job' in
                        project1_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project1-job' in
                         project1_config[0].pipelines['check'].job_list.jobs)
        project2_config = tenant.layout.project_configs.get(
            'review.example.com/org/project2')
        self.assertTrue('common-config-job' in
                        project2_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project2-job' in
                         project2_config[0].pipelines['check'].job_list.jobs)


class TestTenantGroups(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/groups.yaml'

    def test_tenant_groups(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET - set(['project']),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET - set(['project']),
                         tpc.load_classes)
        self.assertTrue('common-config-job' in tenant.layout.jobs)
        self.assertTrue('project1-job' in tenant.layout.jobs)
        self.assertTrue('project2-job' in tenant.layout.jobs)
        project1_config = tenant.layout.project_configs.get(
            'review.example.com/org/project1')
        self.assertTrue('common-config-job' in
                        project1_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project1-job' in
                         project1_config[0].pipelines['check'].job_list.jobs)
        project2_config = tenant.layout.project_configs.get(
            'review.example.com/org/project2')
        self.assertTrue('common-config-job' in
                        project2_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project2-job' in
                         project2_config[0].pipelines['check'].job_list.jobs)


class TestTenantGroups2(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/groups2.yaml'

    def test_tenant_groups2(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2', 'org/project3'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET - set(['project']),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET - set(['project', 'job']),
                         tpc.load_classes)
        self.assertTrue('common-config-job' in tenant.layout.jobs)
        self.assertTrue('project1-job' in tenant.layout.jobs)
        self.assertFalse('project2-job' in tenant.layout.jobs)
        project1_config = tenant.layout.project_configs.get(
            'review.example.com/org/project1')
        self.assertTrue('common-config-job' in
                        project1_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project1-job' in
                         project1_config[0].pipelines['check'].job_list.jobs)
        project2_config = tenant.layout.project_configs.get(
            'review.example.com/org/project2')
        self.assertTrue('common-config-job' in
                        project2_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project2-job' in
                         project2_config[0].pipelines['check'].job_list.jobs)


class TestTenantGroups3(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/groups3.yaml'

    def test_tenant_groups3(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(False, tenant.exclude_unprotected_branches)
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(set(['job']), tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(set(['project', 'job']), tpc.load_classes)
        self.assertTrue('common-config-job' in tenant.layout.jobs)
        self.assertTrue('project1-job' in tenant.layout.jobs)
        self.assertTrue('project2-job' in tenant.layout.jobs)
        project1_config = tenant.layout.project_configs.get(
            'review.example.com/org/project1')
        self.assertTrue('common-config-job' in
                        project1_config[0].pipelines['check'].job_list.jobs)
        self.assertFalse('project1-job' in
                         project1_config[0].pipelines['check'].job_list.jobs)
        project2_config = tenant.layout.project_configs.get(
            'review.example.com/org/project2')
        self.assertTrue('common-config-job' in
                        project2_config[0].pipelines['check'].job_list.jobs)
        self.assertTrue('project2-job' in
                        project2_config[1].pipelines['check'].job_list.jobs)


class TestTenantGroups4(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/groups4.yaml'

    def test_tenant_groups(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(set([]),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(set([]),
                         tpc.load_classes)
        # Check that only one merger:cat job was requested
        # org/project1 and org/project2 have an empty load_classes
        self.assertEqual(1, len(self.merge_job_history.get(MergeRequest.CAT)))
        old_layout = tenant.layout

        # Check that creating a change in project1 doesn't cause a
        # reconfiguration (due to a mistaken belief that we need to
        # load config from it since there is none in memory).
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        new_layout = tenant.layout

        self.assertEqual(old_layout, new_layout)


class TestTenantGroups5(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/groups5.yaml'

    def test_tenant_single_projet_exclude(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(set([]),
                         tpc.load_classes)
        # Check that only one merger:cat job was requested
        # org/project1 and org/project2 have an empty load_classes
        self.assertEqual(1, len(self.merge_job_history.get(MergeRequest.CAT)))


class TestTenantFromScript(TestTenantSimple):
    tenant_config_file = None
    tenant_config_script_file = 'config/tenant-parser/tenant_config_script.py'

    def test_tenant_simple(self):
        TestTenantSimple.test_tenant_simple(self)


class TestTenantUnprotectedBranches(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/unprotected-branches.yaml'

    def test_tenant_unprotected_branches(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(True, tenant.exclude_unprotected_branches)

        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])

        tpc = tenant.project_configs
        project_name = list(tenant.config_projects)[0].canonical_name
        self.assertEqual(False, tpc[project_name].exclude_unprotected_branches)

        project_name = list(tenant.untrusted_projects)[0].canonical_name
        self.assertIsNone(tpc[project_name].exclude_unprotected_branches)

        project_name = list(tenant.untrusted_projects)[1].canonical_name
        self.assertIsNone(tpc[project_name].exclude_unprotected_branches)


class TestTenantIncludeBranches(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/include-branches.yaml'

    def test_tenant_branches(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')

        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])

        tpc = tenant.project_configs
        project_name = list(tenant.config_projects)[0].canonical_name
        self.assertEqual(['master'], tpc[project_name].branches)

        # No branches pass the filter at the start
        project_name = list(tenant.untrusted_projects)[0].canonical_name
        self.assertEqual([], tpc[project_name].branches)

        # Create the foo branch
        self.create_branch('org/project1', 'foo')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project1', 'foo'))
        self.waitUntilSettled()

        # It should pass the filter
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        tpc = tenant.project_configs
        project_name = list(tenant.untrusted_projects)[0].canonical_name
        self.assertEqual(['foo'], tpc[project_name].branches)

        # Create the baz branch
        self.create_branch('org/project1', 'baz')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project1', 'baz'))
        self.waitUntilSettled()

        # It should not pass the filter
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        tpc = tenant.project_configs
        project_name = list(tenant.untrusted_projects)[0].canonical_name
        self.assertEqual(['foo'], tpc[project_name].branches)


class TestTenantExcludeBranches(TestTenantIncludeBranches):
    tenant_config_file = 'config/tenant-parser/exclude-branches.yaml'

    # Same test results as include-branches


class TestTenantExcludeIncludeBranches(TestTenantIncludeBranches):
    tenant_config_file = 'config/tenant-parser/exclude-include-branches.yaml'

    # Same test results as include-branches


class TestTenantExcludeAll(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/exclude-all.yaml'

    def test_tenant_exclude_all(self):
        """
        Tests that excluding all configuration of project1 in tenant-one
        doesn't remove the configuration of project1 in tenant-two.
        """
        # The config in org/project5 depends on config in org/project1 so
        # validate that there are no config errors in that tenant.
        tenant_two = self.scheds.first.sched.abide.tenants.get('tenant-two')
        self.assertEqual(
            len(tenant_two.layout.loading_errors), 0,
            "No error should have been accumulated")


class TestTenantConfigBranches(ZuulTestCase):
    tenant_config_file = 'config/tenant-parser/simple.yaml'

    def _validate_job(self, job, branch):
        tenant_one = self.scheds.first.sched.abide.tenants.get('tenant-one')
        jobs = tenant_one.layout.getJobs(job)
        self.assertEqual(len(jobs), 1)
        self.assertIn(jobs[0].source_context.branch, branch)

    def test_tenant_config_load_branch(self):
        """
        Tests that when specifying branches for a project only those branches
        are parsed.
        """
        # Job must be defined in master
        common_job = 'common-config-job'
        self._validate_job(common_job, 'master')

        self.log.debug('Creating branches')
        self.create_branch('common-config', 'stable')
        self.create_branch('common-config', 'feat_x')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'common-config', 'stable'))
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'common-config', 'feat_x'))
        self.waitUntilSettled()

        # Job must be defined in master
        self._validate_job(common_job, 'master')

        # Reconfigure with load-branch stable for common-config
        self.newTenantConfig('config/tenant-parser/branch.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))

        # Now job must be defined on stable branch
        self._validate_job(common_job, 'stable')

        # Now try to break the config in common-config on stable
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: base
                parent: non-existing
            """)
        file_dict = {'zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('common-config', 'stable', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # No job should have run due to the change introducing a config error
        self.assertHistory([])
        self.assertTrue(A.reported)
        self.assertTrue('Job non-existing not defined' in A.messages[0])


class TestSplitConfig(ZuulTestCase):
    tenant_config_file = 'config/split-config/main.yaml'

    def test_split_config(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertIn('project-test1', tenant.layout.jobs)
        self.assertIn('project-test2', tenant.layout.jobs)
        test1 = tenant.layout.getJob('project-test1')
        self.assertEqual(test1.source_context.project_name, 'common-config')
        self.assertEqual(test1.source_context.branch, 'master')
        self.assertEqual(test1.source_context.path, 'zuul.d/jobs.yaml')
        self.assertTrue(
            tenant.isTrusted(test1.source_context.project_canonical_name),
            True)
        test2 = tenant.layout.getJob('project-test2')
        self.assertEqual(test2.source_context.project_name, 'common-config')
        self.assertEqual(test2.source_context.branch, 'master')
        self.assertEqual(test2.source_context.path, 'zuul.d/more-jobs.yaml')
        self.assertTrue(
            tenant.isTrusted(test2.source_context.project_canonical_name),
            True)

        self.assertNotEqual(test1.source_context, test2.source_context)
        self.assertTrue(test1.source_context.isSameProject(
            test2.source_context))

        project_config = tenant.layout.project_configs.get(
            'review.example.com/org/project')
        self.assertIn('project-test1',
                      project_config[0].pipelines['check'].job_list.jobs)
        project1_config = tenant.layout.project_configs.get(
            'review.example.com/org/project1')
        self.assertIn('project1-project2-integration',
                      project1_config[0].pipelines['check'].job_list.jobs)

        # This check ensures the .zuul.ignore flag file is working in
        # the config directory.
        self.assertEqual(
            len(tenant.layout.loading_errors), 0)

    def test_dynamic_split_config(self):
        in_repo_conf = textwrap.dedent(
            """
            - project:
                name: org/project1
                check:
                  jobs:
                    - project-test1
            """)
        file_dict = {'.zuul.d/gate.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        # project1-project2-integration test removed, only want project-test1
        self.assertHistory([
            dict(name='project-test1', result='SUCCESS', changes='1,1')])

    def test_config_path_conflict(self):
        def add_file(project, path):
            new_file = textwrap.dedent(
                """
                - job:
                    name: test-job
                """
            )
            file_dict = {path: new_file}
            A = self.fake_gerrit.addFakeChange(project, 'master', 'A',
                                               files=file_dict)
            self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
            self.waitUntilSettled()
            return A

        log_fixture = self.useFixture(
            fixtures.FakeLogger(level=logging.WARNING))

        log_fixture._output.truncate(0)
        A = add_file("common-config", "zuul.yaml")
        self.assertIn("Configuration in common-config/zuul.d/jobs.yaml@master "
                      "ignored because project-branch is already configured",
                      log_fixture.output)
        self.assertIn("Configuration in common-config/zuul.d/jobs.yaml@master "
                      "ignored because project-branch is already configured",
                      A.messages[0])

        log_fixture._output.truncate(0)
        add_file("org/project1", ".zuul.yaml")
        self.assertIn("Configuration in org/project1/.zuul.d/gate.yaml@master "
                      "ignored because project-branch is already configured",
                      log_fixture.output)


class TestConfigConflict(ZuulTestCase):
    tenant_config_file = 'config/conflict-config/main.yaml'

    def test_conflict_config(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        jobs = sorted(tenant.layout.jobs.keys())
        self.assertEqual(
            ['base', 'noop', 'trusted-zuul.yaml-job',
             'untrusted-zuul.yaml-job'],
            jobs)

        tenant = self.scheds.first.sched.abide.tenants.get("tenant-one")
        errors = tenant.layout.loading_errors
        self.assertEqual(len(errors), 4)

        for error in errors:
            self.assertEqual(error.severity, model.SEVERITY_WARNING)
            self.assertEqual(error.name, 'Multiple Project Configurations')


class TestUnparsedConfigCache(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_config_caching(self):
        sched = self.scheds.first.sched
        cache = sched.unparsed_config_cache
        tenant = sched.abide.tenants["tenant-one"]

        common_cache = cache.getFilesCache("review.example.com/common-config",
                                           "master")
        object_common_cache = sched.abide.getConfigObjectCache(
            "review.example.com/common-config", "master")
        tpc = tenant.project_configs["review.example.com/common-config"]
        self.assertTrue(common_cache.isValidFor(tpc, min_ltime=-1))
        self.assertEqual(len(common_cache), 1)
        self.assertIn("zuul.yaml", common_cache)
        self.assertTrue(len(common_cache["zuul.yaml"]) > 0)
        self.assertEqual(object_common_cache.entries['zuul.yaml'].ltime,
                         common_cache.ltime)

        project_cache = cache.getFilesCache("review.example.com/org/project",
                                            "master")
        object_project_cache = sched.abide.getConfigObjectCache(
            "review.example.com/org/project", "master")
        # Cache of org/project should be valid but empty (no in-repo config)
        tpc = tenant.project_configs["review.example.com/org/project"]
        self.assertTrue(project_cache.isValidFor(tpc, min_ltime=-1))
        self.assertEqual(len(project_cache), 0)
        self.assertEqual(object_project_cache.entries['zuul.yaml'].ltime,
                         project_cache.ltime)

    @okay_tracebacks('cannot schedule new futures after shutdown')
    def test_cache_use(self):
        sched = self.scheds.first.sched
        # Stop cleanup jobs so it's not removing projects from
        # the cache during the test.
        sched.apsched.shutdown()
        tenant = sched.abide.tenants['tenant-one']
        _, project = tenant.getProject('org/project2')

        cache = self.scheds.first.sched.unparsed_config_cache
        files_cache = cache.getFilesCache(
            "review.example.com/org/project2", "master")
        zk_initial_ltime = files_cache.ltime
        config_object_cache = sched.abide.getConfigObjectCache(
            "review.example.com/org/project2", "master")
        self.assertEqual(zk_initial_ltime,
                         config_object_cache.entries['zuul.yaml'].ltime)

        # Get the current ltime from Zookeeper and run a full reconfiguration,
        # so that we know all items in the cache have a larger ltime.
        ltime = self.zk_client.getCurrentLtime()
        self.scheds.first.fullReconfigure()
        self.assertGreater(files_cache.ltime, zk_initial_ltime)
        config_object_cache = sched.abide.getConfigObjectCache(
            "review.example.com/org/project2", "master")
        self.assertEqual(files_cache.ltime,
                         config_object_cache.entries['zuul.yaml'].ltime)

        # Clear the unparsed branch cache so all projects (except for
        # org/project2) are retrieved from the cache in Zookeeper.
        sched.abide.config_object_cache.clear()
        del self.merge_job_history

        # Create a tenant reconfiguration event with a known ltime that is
        # smaller than the ltime of the items in the cache.
        event = model.TenantReconfigureEvent(
            tenant.name, project.canonical_name, branch_name=None)
        event.zuul_event_ltime = ltime
        sched.management_events[tenant.name].put(event, needs_result=False)
        self.waitUntilSettled()

        # As the cache should be valid (cache ltime of org/project2 newer than
        # event ltime) we don't expect any cat jobs.
        self.assertIsNone(self.merge_job_history.get(MergeRequest.CAT))

        # Set canary value so we can detect if the configloader used
        # the cache in Zookeeper (it shouldn't).
        common_cache = cache.getFilesCache("review.example.com/common-config",
                                           "master")
        common_cache.setValidFor({"CANARY"}, set(), common_cache.ltime)
        del self.merge_job_history

        # Create a tenant reconfiguration event with a known ltime that is
        # smaller than the ltime of the items in the cache.
        event = model.TenantReconfigureEvent(
            tenant.name, project.canonical_name, branch_name=None)
        event.zuul_event_ltime = ltime
        sched.management_events[tenant.name].put(event, needs_result=False)
        self.waitUntilSettled()

        config_object_cache = sched.abide.getConfigObjectCache(
            "review.example.com/common-config", "master")
        self.assertEqual(common_cache.ltime,
                         config_object_cache.entries['zuul.yaml'].ltime)
        self.assertNotIn("CANARY", config_object_cache.entries)

        # As the cache should be valid (cache ltime of org/project2 newer than
        # event ltime) we don't expect any cat jobs.
        self.assertIsNone(self.merge_job_history.get(MergeRequest.CAT))
        sched.apsched.start()


class TestAuthorizationRuleParser(ZuulTestCase):
    tenant_config_file = 'config/tenant-parser/authorizations.yaml'

    def test_rules_are_loaded(self):
        rules = self.scheds.first.sched.abide.authz_rules
        self.assertTrue('auth-rule-one' in rules,
                        self.scheds.first.sched.abide)
        self.assertTrue('auth-rule-two' in rules,
                        self.scheds.first.sched.abide)
        claims_1 = {'sub': 'venkman'}
        claims_2 = {'sub': 'gozer',
                    'iss': 'another_dimension'}
        self.assertTrue(rules['auth-rule-one'](claims_1))
        self.assertTrue(not rules['auth-rule-one'](claims_2))
        self.assertTrue(not rules['auth-rule-two'](claims_1))
        self.assertTrue(rules['auth-rule-two'](claims_2))

    def test_parse_simplest_rule_from_yaml(self):
        rule_d = {'name': 'my-rule',
                  'conditions': {'sub': 'user1'}
                 }
        rule = AuthorizationRuleParser().fromYaml(rule_d)
        self.assertEqual('my-rule', rule.name)
        claims = {'iss': 'my-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbusters']}
        self.assertTrue(rule(claims))
        claims = {'iss': 'my-2nd-idp',
                  'sub': 'user2',
                  'groups': ['admin', 'ghostbusters']}
        self.assertFalse(rule(claims))

    def test_parse_AND_rule_from_yaml(self):
        rule_d = {'name': 'my-rule',
                  'conditions': {'sub': 'user1',
                                 'iss': 'my-idp'}
                 }
        rule = AuthorizationRuleParser().fromYaml(rule_d)
        self.assertEqual('my-rule', rule.name)
        claims = {'iss': 'my-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbusters']}
        self.assertTrue(rule(claims))
        claims = {'iss': 'my-2nd-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbusters']}
        self.assertFalse(rule(claims))

    def test_parse_OR_rule_from_yaml(self):
        rule_d = {'name': 'my-rule',
                  'conditions': [{'sub': 'user1',
                                  'iss': 'my-idp'},
                                 {'sub': 'user2',
                                  'iss': 'my-2nd-idp'}
                                ]
                 }
        rule = AuthorizationRuleParser().fromYaml(rule_d)
        self.assertEqual('my-rule', rule.name)
        claims = {'iss': 'my-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbusters']}
        self.assertTrue(rule(claims))
        claims = {'iss': 'my-2nd-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbusters']}
        self.assertFalse(rule(claims))
        claims = {'iss': 'my-2nd-idp',
                  'sub': 'user2',
                  'groups': ['admin', 'ghostbusters']}
        self.assertTrue(rule(claims))

    def test_parse_rule_with_list_claim_from_yaml(self):
        rule_d = {'name': 'my-rule',
                  'conditions': [{'groups': 'ghostbusters',
                                  'iss': 'my-idp'},
                                 {'sub': 'user2',
                                  'iss': 'my-2nd-idp'}
                                ],
                 }
        rule = AuthorizationRuleParser().fromYaml(rule_d)
        self.assertEqual('my-rule', rule.name)
        claims = {'iss': 'my-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbusters']}
        self.assertTrue(rule(claims))
        claims = {'iss': 'my-idp',
                  'sub': 'user1',
                  'groups': ['admin', 'ghostbeaters']}
        self.assertFalse(rule(claims))
        claims = {'iss': 'my-2nd-idp',
                  'sub': 'user2',
                  'groups': ['admin', 'ghostbusters']}
        self.assertTrue(rule(claims))

    def test_check_complex_rule_from_yaml_jsonpath(self):
        rule_d = {'name': 'my-rule',
                  'conditions': [{'hello.this.is': 'a complex value'},
                                ],
                 }
        rule = AuthorizationRuleParser().fromYaml(rule_d)
        self.assertEqual('my-rule', rule.name)
        claims = {'iss': 'my-idp',
                  'hello': {
                      'this': {
                          'is': 'a complex value'
                      },
                      'and': {
                          'this one': 'too'
                      }
                  }
                 }
        self.assertTrue(rule(claims))

    def test_check_complex_rule_from_yaml_nested_dict(self):
        rule_d = {'name': 'my-rule',
                  'conditions': [{'hello': {'this': {'is': 'a complex value'
                                                    }
                                           }
                                 },
                                ],
                 }
        rule = AuthorizationRuleParser().fromYaml(rule_d)
        self.assertEqual('my-rule', rule.name)
        claims = {'iss': 'my-idp',
                  'hello': {
                      'this': {
                          'is': 'a complex value'
                      },
                      'and': {
                          'this one': 'too'
                      }
                  }
                 }
        self.assertTrue(rule(claims))


class TestAuthorizationRuleParserWithTemplating(ZuulTestCase):
    tenant_config_file = 'config/tenant-parser/authorizations-templating.yaml'

    def test_rules_are_loaded(self):
        rules = self.scheds.first.sched.abide.authz_rules
        self.assertTrue('tenant-admin' in rules, self.scheds.first.sched.abide)
        self.assertTrue('tenant-admin-complex' in rules,
                        self.scheds.first.sched.abide)

    def test_tenant_substitution(self):
        claims_1 = {'group': 'tenant-one-admin'}
        claims_2 = {'group': 'tenant-two-admin'}
        rules = self.scheds.first.sched.abide.authz_rules
        tenant_one = self.scheds.first.sched.abide.tenants.get('tenant-one')
        tenant_two = self.scheds.first.sched.abide.tenants.get('tenant-two')
        self.assertTrue(rules['tenant-admin'](claims_1, tenant_one))
        self.assertTrue(rules['tenant-admin'](claims_2, tenant_two))
        self.assertTrue(not rules['tenant-admin'](claims_1, tenant_two))
        self.assertTrue(not rules['tenant-admin'](claims_2, tenant_one))

    def test_tenant_substitution_in_list(self):
        claims_1 = {'group': ['tenant-one-admin', 'some-other-tenant']}
        claims_2 = {'group': ['tenant-two-admin', 'some-other-tenant']}
        rules = self.scheds.first.sched.abide.authz_rules
        tenant_one = self.scheds.first.sched.abide.tenants.get('tenant-one')
        tenant_two = self.scheds.first.sched.abide.tenants.get('tenant-two')
        self.assertTrue(rules['tenant-admin'](claims_1, tenant_one))
        self.assertTrue(rules['tenant-admin'](claims_2, tenant_two))
        self.assertTrue(not rules['tenant-admin'](claims_1, tenant_two))
        self.assertTrue(not rules['tenant-admin'](claims_2, tenant_one))

    def test_tenant_substitution_in_dict(self):
        claims_2 = {
            'path': {
                'to': {
                    'group': 'tenant-two-admin'
                }
            }
        }
        rules = self.scheds.first.sched.abide.authz_rules
        tenant_one = self.scheds.first.sched.abide.tenants.get('tenant-one')
        tenant_two = self.scheds.first.sched.abide.tenants.get('tenant-two')
        self.assertTrue(not rules['tenant-admin-complex'](claims_2,
                                                          tenant_one))
        self.assertTrue(rules['tenant-admin-complex'](claims_2, tenant_two))


class TestTenantExtra(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/extra.yaml'

    def test_tenant_extra(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertTrue('project2-extra-file' in tenant.layout.jobs)
        self.assertTrue('project2-extra-dir' in tenant.layout.jobs)
        self.assertTrue('project6-extra-dir' in tenant.layout.jobs)

    def test_dynamic_extra(self):
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: project2-extra-file2
                parent: common-config-job
            - project:
                name: org/project2
                check:
                  jobs:
                    - project2-extra-file2
            """)
        file_dict = {'extra.yaml': in_repo_conf, '.zuul.yaml': ''}
        A = self.fake_gerrit.addFakeChange('org/project2', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='common-config-job', result='SUCCESS', changes='1,1'),
            dict(name='project2-extra-file2', result='SUCCESS', changes='1,1'),
        ], ordered=False)

    def test_dynamic_extra_dir(self):
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: project6-extra-dir2
                parent: common-config-job
            - project:
                check:
                  jobs:
                    - project6-extra-dir
                    - project6-extra-dir2
            """)
        file_dict = {
            'other/extra.d/new/extra.yaml': in_repo_conf,
        }
        A = self.fake_gerrit.addFakeChange('org/project6', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project6-extra-dir', result='SUCCESS', changes='1,1'),
            dict(name='project6-extra-dir2', result='SUCCESS', changes='1,1'),
        ], ordered=False)

    def test_extra_reconfigure(self):
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: project2-extra-file2
                parent: common-config-job
            - project:
                name: org/project2
                check:
                  jobs:
                    - project2-extra-file2
            """)
        file_dict = {'extra.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project2', 'master', 'A',
                                           files=file_dict)
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.waitUntilSettled()

        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='common-config-job', result='SUCCESS', changes='2,1'),
            dict(name='project2-job', result='SUCCESS', changes='2,1'),
            dict(name='project2-extra-file2', result='SUCCESS', changes='2,1'),
        ], ordered=False)


class TestTenantExtraConfigsInvalidType(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/extra_invalid_type.yaml'

    # This test raises a config error during the startup of the test
    # case which makes the first scheduler fail during its startup.
    # The second (or any additional) scheduler won't even run as the
    # startup is serialized in tests/base.py.
    # Thus it doesn't make sense to execute this test with multiple
    # schedulers.
    scheduler_count = 1

    def setUp(self):
        err = "Expected str or list of str for extra-config-paths.*"
        with testtools.ExpectedException(vs.MultipleInvalid, err):
            super().setUp()

    def test_tenant_extra_configs_invalid_type(self):
        # The magic is in setUp
        pass


class TestTenantExtraConfigsInvalidValue(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/extra_invalid_value.yaml'

    # This test raises a config error during the startup of the test
    # case which makes the first scheduler fail during its startup.
    # The second (or any additional) scheduler won't even run as the
    # startup is serialized in tests/base.py.
    # Thus it doesn't make sense to execute this test with multiple
    # schedulers.
    scheduler_count = 1

    def setUp(self):
        err = "Default zuul configs are not allowed in extra-config-paths.*"
        with testtools.ExpectedException(vs.MultipleInvalid, err):
            super().setUp()

    def test_tenant_extra_configs_invalid_value(self):
        # The magic is in setUp
        pass


class TestTenantDuplicate(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/duplicate.yaml'

    # This test raises a config error during the startup of the test
    # case which makes the first scheduler fail during its startup.
    # The second (or any additional) scheduler won't even run as the
    # startup is serialized in tests/base.py.
    # Thus it doesn't make sense to execute this test with multiple
    # schedulers.
    scheduler_count = 1

    def setUp(self):
        with testtools.ExpectedException(Exception, 'Duplicate configuration'):
            super().setUp()

    def test_tenant_dupe(self):
        # The magic is in setUp
        pass


class TestTenantTPCLoadingFailure(TenantParserTestCase):
    tenant_config_file = ('config/tenant-parser/'
                          'simple.yaml')
    scheduler_count = 1

    @okay_tracebacks('may not configure')
    def test_tenant_tpc_loading_failure(self):
        # This tests the behavior when we fail to load TPCs,
        # especially after a previously successful loading.  If we
        # fail to fully load information about a tenant, it can be
        # dangerous to proceed with run loop or cleanup handling in
        # the scheduler.

        # The test starts with a valid configuration.  This next
        # configuration raises an error during TPC loading
        self.newTenantConfig('config/tenant-parser/'
                             'superproject-config-project.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        # Counter-intuitively, the tenant *is* up to date on this
        # scheduler.  We are running with the previous configuration
        # since the new one didn't load.
        self.assertTrue(
            self.scheds.first.sched.isTenantLayoutUpToDate('tenant-one')
        )

        # Start a second scheduler which will not have the benefit of
        # a config in memory.
        app = self.createScheduler()
        app.start()
        self.assertEqual(len(self.scheds), 2)

        # In this case, the layout is not up to date since we don't
        # have anything to fall back on.
        self.assertFalse(
            app.sched.isTenantLayoutUpToDate('tenant-one')
        )


class TestTenantSuperprojectConfigProject(TenantParserTestCase):
    tenant_config_file = ('config/tenant-parser/'
                          'superproject-config-project.yaml')
    scheduler_count = 1

    def setUp(self):
        # Test that we get an error trying to configure a
        # config-project
        err = ".*may not configure config-project.*"
        with testtools.ExpectedException(Exception, err):
            super().setUp()

    def test_tenant_superproject_config_project(self):
        # The magic is in setUp
        pass


class TestTenantSuperprojectConfigProjectRegex(TenantParserTestCase):
    tenant_config_file = ('config/tenant-parser/'
                          'superproject-config-project-regex.yaml')
    scheduler_count = 1

    def setUp(self):
        # Test that we get an error trying to configure a
        # config-project via regex
        err = ".*may not configure config-project.*"
        with testtools.ExpectedException(Exception, err):
            super().setUp()

    def test_tenant_superproject_config_project_regex(self):
        # The magic is in setUp
        pass


class TestTenantConfigSuperproject(TenantParserTestCase):
    tenant_config_file = ('config/tenant-parser/'
                          'config-superproject.yaml')
    scheduler_count = 1

    def setUp(self):
        # Test that we get an error trying to use configure-projects
        # on a config-project
        with testtools.ExpectedException(vs.MultipleInvalid):
            super().setUp()

    def test_tenant_config_superproject(self):
        # The magic is in setUp
        pass


class TestTenantProvider(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/provider.yaml'

    def test_tenant_provider(self):
        # Test config/untrusted defaults
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET, tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET,
                         tpc.load_classes)


class TestTenantProvider2(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/provider2.yaml'

    def test_tenant_provider2(self):
        # Test the include-provider-config setting, and test it with
        # "exclude".
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.CONFIG_SET - set(['semaphore']),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.UNTRUSTED_SET | self.PROVIDER_SET,
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(
            (self.UNTRUSTED_SET | self.PROVIDER_SET) - set(['project']),
            tpc.load_classes)


class TestTenantProvider3(TenantParserTestCase):
    tenant_config_file = 'config/tenant-parser/provider3.yaml'

    def test_tenant_provider3(self):
        # Test the include-provider-config setting, and test it with
        # "include".
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(['common-config'],
                         [x.name for x in tenant.config_projects])
        self.assertEqual(['org/project1', 'org/project2'],
                         [x.name for x in tenant.untrusted_projects])
        project = list(tenant.config_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.PROVIDER_SET | set(['pipeline']),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[0]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(self.PROVIDER_SET | set(['job']),
                         tpc.load_classes)
        project = list(tenant.untrusted_projects)[1]
        tpc = tenant.project_configs[project.canonical_name]
        self.assertEqual(
            self.PROVIDER_SET | set(['project']),
            tpc.load_classes)


class TestMergeMode(ZuulTestCase):
    config_file = 'zuul-connections-gerrit-and-github.conf'

    def _test_default_merge_mode(self, driver_default, host):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        md = layout.getProjectMetadata(
            f'{host}/org/project-empty')
        self.assertEqual(driver_default, md.merge_mode)
        md = layout.getProjectMetadata(
            f'{host}/org/regex-empty-project-empty')
        self.assertEqual(driver_default, md.merge_mode)
        md = layout.getProjectMetadata(
            f'{host}/org/regex-empty-project-squash')
        self.assertEqual(model.MERGER_SQUASH_MERGE, md.merge_mode)
        md = layout.getProjectMetadata(
            f'{host}/org/regex-cherry-project-empty')
        self.assertEqual(model.MERGER_CHERRY_PICK, md.merge_mode)
        md = layout.getProjectMetadata(
            f'{host}/org/regex-cherry-project-squash')
        self.assertEqual(model.MERGER_SQUASH_MERGE, md.merge_mode)

    @simple_layout('layouts/merge-mode-default.yaml')
    def test_default_merge_mode_gerrit(self):
        self._test_default_merge_mode(model.MERGER_MERGE_RESOLVE,
                                      'review.example.com')

    @simple_layout('layouts/merge-mode-default.yaml', driver='github')
    def test_default_merge_mode_github(self):
        self._test_default_merge_mode(model.MERGER_MERGE_ORT,
                                      'github.com')


class TestDefaultBranch(ZuulTestCase):
    config_file = 'zuul-connections-gerrit-and-github.conf'

    @simple_layout('layouts/default-branch.yaml')
    def test_default_branch(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        md = layout.getProjectMetadata(
            'review.example.com/org/project-default')
        self.assertEqual('master', md.default_branch)
        md = layout.getProjectMetadata(
            'review.example.com/org/regex-default-project-empty')
        self.assertEqual('master', md.default_branch)
        md = layout.getProjectMetadata(
            'review.example.com/org/regex-default-project-develop')
        self.assertEqual('develop', md.default_branch)
        md = layout.getProjectMetadata(
            'review.example.com/org/regex-override-project-empty')
        self.assertEqual('regex', md.default_branch)
        md = layout.getProjectMetadata(
            'review.example.com/org/regex-override-project-develop')
        self.assertEqual('develop', md.default_branch)

    @simple_layout('layouts/default-branch.yaml', driver='github')
    def test_default_branch_upstream(self):
        self.waitUntilSettled()
        github = self.fake_github.getGithubClient()
        repo = github.repo_from_project('org/project-default')
        repo._repodata['default_branch'] = 'foobar'
        connection = self.scheds.first.connections.connections['github']
        connection.clearBranchCache()
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        md = layout.getProjectMetadata(
            'github.com/org/project-default')
        self.assertEqual('foobar', md.default_branch)
        md = layout.getProjectMetadata(
            'github.com/org/regex-default-project-empty')
        self.assertEqual('master', md.default_branch)
        md = layout.getProjectMetadata(
            'github.com/org/regex-default-project-develop')
        self.assertEqual('develop', md.default_branch)
        md = layout.getProjectMetadata(
            'github.com/org/regex-override-project-empty')
        self.assertEqual('regex', md.default_branch)
        md = layout.getProjectMetadata(
            'github.com/org/regex-override-project-develop')
        self.assertEqual('develop', md.default_branch)


class TestNodepoolConfig(LauncherBaseTestCase):
    config_file = 'zuul-connections-nodepool.conf'

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_nodepool_config(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        self.assertEqual(1, len(layout.images))
        image = layout.images['debian']
        self.assertEqual('debian', image.name)
        self.assertEqual('cloud', image.type)
        flavor = layout.flavors['normal']
        self.assertEqual('normal', flavor.name)
        label = layout.labels['debian-normal']
        self.assertEqual('debian-normal', label.name)
        self.assertEqual('debian', label.image)
        self.assertEqual('normal', label.flavor)
        label = layout.labels['debian-dedicated']
        self.assertEqual('debian-dedicated', label.name)
        self.assertEqual('debian', label.image)
        self.assertEqual('dedicated', label.flavor)
        label = layout.labels['debian-invalid']
        self.assertEqual('debian-invalid', label.name)
        self.assertEqual('debian', label.image)
        self.assertEqual('invalid', label.flavor)
        section = layout.sections['aws-base']
        self.assertEqual('aws-base', section.name)
        self.assertEqual(True, section.abstract)
        self.assertTrue('launch-timeout' in section.config)
        provider_config = layout.provider_configs['aws-us-east-1-main']
        self.assertEqual('aws-us-east-1-main', provider_config.name)
        self.assertEqual('aws-us-east-1', provider_config.section)
        provider = layout.providers['aws-us-east-1-main']
        self.assertEqual(4, len(provider.labels))
        labels = sorted([x for x in provider.labels.keys()])
        self.assertEqual('debian-dedicated', labels[0])
        self.assertEqual('debian-invalid', labels[1])
        self.assertEqual('debian-large', labels[2])
        self.assertEqual('debian-normal', labels[3])

    @simple_layout('layouts/nodepool-defaults.yaml', enable_nodepool=True)
    def test_nodepool_config_defaults(self):
        # Some notes about the current inheritance scheme:
        # 1) Providers inherit attributes from sections (and so on)
        #   A) Dictionaries are merged
        #   B) Flavors, images, labels lists behave as units
        #   C) Their defaults are applied at each inheritance level
        #   D) Later definitions completely override earlier ones
        #      (are not merged)
        # 2) Labels inherit some attributes from their images and flavors
        #   A) Labels override flavors override images
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout

        self.assertEqual(2, len(layout.images))
        image = layout.images['debian']
        self.assertEqual('debian', image.name)
        self.assertEqual('cloud', image.type)
        image = layout.images['debian-local']
        self.assertEqual('debian-local', image.name)
        self.assertEqual('zuul', image.type)

        self.assertEqual(2, len(layout.flavors))
        flavor = layout.flavors['normal']
        self.assertEqual('normal', flavor.name)
        flavor = layout.flavors['large']
        self.assertEqual('large', flavor.name)

        self.assertEqual(3, len(layout.labels))
        label = layout.labels['debian-normal']
        self.assertEqual('debian-normal', label.name)
        self.assertEqual('debian', label.image)
        self.assertEqual('normal', label.flavor)

        label = layout.labels['debian-large']
        self.assertEqual('debian-large', label.name)
        self.assertEqual('debian', label.image)
        self.assertEqual('large', label.flavor)
        label = layout.labels['debian-local-normal']
        self.assertEqual('debian-local-normal', label.name)
        self.assertEqual('debian-local', label.image)
        self.assertEqual('normal', label.flavor)

        section = layout.sections['aws-base']
        self.assertEqual('aws-base', section.name)
        self.assertEqual(True, section.abstract)
        self.assertTrue('launch-timeout' in section.config)
        provider_config = layout.provider_configs['aws-us-east-1-main']
        self.assertEqual('aws-us-east-1-main', provider_config.name)
        self.assertEqual('aws-us-east-1', provider_config.section)
        provider = layout.providers['aws-us-east-1-main']
        self.assertEqual(3, len(provider.labels))
        labels = sorted([x for x in provider.labels.keys()])
        self.assertEqual('debian-large', labels[0])
        self.assertEqual('debian-local-normal', labels[1])
        self.assertEqual('debian-normal', labels[2])

        # Check the inherited values
        label = provider.labels['debian-normal']
        self.assertEqual('gp2', label.volume_type)
        self.assertEqual(10, label.volume_size)
        self.assertEqual(40, label.iops)
        self.assertEqual(20, label.throughput)
        self.assertEqual(900, label.max_ready_age)
        self.assertEqual('required', label.imds_http_tokens)
        self.assertFalse(label.host_key_checking)
        self.assertEqual(['sg1'], label.security_group_ids)

        label = provider.labels['debian-large']
        self.assertEqual('gp3', label.volume_type)
        self.assertEqual(10, label.volume_size)
        self.assertEqual(70, label.iops)
        self.assertEqual(20, label.throughput)
        self.assertEqual(300, label.max_ready_age)
        self.assertEqual('required', label.imds_http_tokens)
        self.assertFalse(label.host_key_checking)
        self.assertEqual(['sg1'], label.security_group_ids)

        label = provider.labels['debian-local-normal']
        self.assertEqual('gp3', label.volume_type)
        self.assertEqual(10, label.volume_size)
        self.assertEqual(40, label.iops)
        self.assertEqual(60, label.throughput)
        self.assertEqual(900, label.max_ready_age)
        self.assertEqual('required', label.imds_http_tokens)
        self.assertFalse(label.host_key_checking)
        self.assertEqual(['sg1'], label.security_group_ids)

    @simple_layout('layouts/nodepool-defaults-redefinition.yaml',
                   enable_nodepool=True)
    def test_nodepool_config_defaults_redefinition(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        provider = layout.providers['aws-us-east-1-main']
        self.assertEqual(3, len(provider.labels))

        # Check the inherited values
        label = provider.labels['debian-normal']
        self.assertEqual(40, label.iops)

        label = provider.labels['debian-large']
        self.assertEqual(70, label.iops)

        label = provider.labels['debian-local-normal']
        self.assertEqual(40, label.iops)

    @simple_layout('layouts/nodepool-defaults-redefinition-final.yaml',
                   enable_nodepool=True)
    def test_nodepool_config_defaults_redefinition_final(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        self.assertEqual(1, len(layout.loading_errors))
        self.assertIn('redefine final flavor', layout.loading_errors[0].error)

    @simple_layout('layouts/nodepool-defaults-inheritance.yaml',
                   enable_nodepool=True)
    def test_nodepool_config_defaults_inheritance(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout

        provider = layout.providers['aws-us-east-1-main']
        self.assertEqual(2, len(provider.labels))

        # Check the inherited values
        label = provider.labels['debian-normal']
        self.assertEqual(30, label.iops)

        label = provider.labels['debian-local-normal']
        self.assertEqual(30, label.iops)

    @simple_layout('layouts/nodepool-defaults-inheritance-final.yaml',
                   enable_nodepool=True)
    def test_nodepool_config_defaults_inheritance_final(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout

        provider = layout.providers['aws-us-east-1-main']
        self.assertEqual(2, len(provider.labels))

        # Check the inherited values
        label = provider.labels['debian-normal']
        self.assertEqual(20, label.iops)

        label = provider.labels['debian-local-normal']
        self.assertEqual(30, label.iops)

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_section_inheritance(self):
        # Verify that a section may inherit from a section in a
        # different project.

        in_repo_conf = textwrap.dedent(
            """
            - section:
                name: newsection
                parent: aws-base
            - provider:
                name: newprovider
                section: newsection
                region: us-west-2
            """)
        file_dict = {'zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(A.reported, 1)
        self.assertEqual(A.patchsets[-1]['approvals'][0]['value'], '1')
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertTrue('newprovider' in tenant.layout.providers)

    @simple_layout('layouts/nodepool-provider-flavors.yaml',
                   enable_nodepool=True)
    def test_nodepool_provider_flavors(self):
        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        self.assertEqual(1, len(layout.loading_errors))
        self.assertIn('Flavor configuration not permitted in',
                      layout.loading_errors[0].error)



# Copyright 2021-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import cachetools.func
import json
import logging
import math
import random
import string
from concurrent.futures import ThreadPoolExecutor

from . import azul

from zuul import exceptions
from zuul.driver.azure.azuremodel import AzureResource, AzureInstance
from zuul.driver.util import (
    ImageUploader,
    LazyExecutorTTLCache,
    RateLimiter,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseImageUploadJob,
    BaseProviderEndpoint,
    statemachine
)

CACHE_TTL = 10
MIB = 1024 ** 2


def quota_info_from_sku(sku, spot=False):
    if not sku:
        return QuotaInformation(instances=1)

    cores = 0
    spot_cores = 0
    ram = 0
    for cap in sku['capabilities']:
        if cap['name'] == 'vCPUs':
            if spot:
                spot_cores = int(cap['value'])
            else:
                cores = int(cap['value'])
        if cap['name'] == 'MemoryGB':
            ram = int(float(cap['value']) * 1024)
    return QuotaInformation(
        cores=cores,
        spot_cores=spot_cores,
        ram=ram,
        instances=1)


def generate_password():
    while True:
        chars = random.choices(string.ascii_lowercase +
                               string.ascii_uppercase +
                               string.digits,
                               k=64)
        if ((set(string.ascii_lowercase) & set(chars)) and
            (set(string.ascii_uppercase) & set(chars)) and
            (set(string.digits) & set(chars))):
            return ''.join(chars)


class AzureDeleteStateMachine(statemachine.StateMachine):
    VM_DELETING = 'deleting vm'
    COMPLETE = 'complete'

    def __init__(self, endpoint, node, log):
        self.log = log
        self.endpoint = endpoint
        self.node = node
        super().__init__(node.delete_state)
        self.vm = None

    def advance(self):
        if self.state == self.START:
            if self.vm is None:
                self.vm = self.endpoint._deleteVirtualMachine(
                    self.node.azure_vm_id)
            self.state = self.VM_DELETING

        if self.state == self.VM_DELETING:
            self.vm = self.endpoint._refreshDelete(self.vm)
            if self.vm is None:
                self.state = self.COMPLETE

        if self.state == self.COMPLETE:
            self.complete = True


class AzureCreateStateMachine(statemachine.StateMachine):
    VM_CREATING_SUBMIT = 'submit creating vm'
    VM_CREATING = 'creating vm'
    NIC_QUERY = 'querying nic'
    PIP_QUERY = 'querying pip'
    COMPLETE = 'complete'

    def __init__(self, provider, endpoint, node, hostname, label,
                 flavor, image, image_external_id, tags, log):
        self.log = log
        self.provider = provider
        self.resource_group = provider.resource_group
        self.endpoint = endpoint
        self.node = node
        self.tags = tags.copy()
        self.hostname = hostname
        self.label = label
        self.flavor = flavor
        self.image = image
        self.public_ipv4 = None
        self.public_ipv6 = None
        self.nic = None
        self.create_future = None
        super().__init__(node.create_state)
        self.attempts = node.create_state.get("attempts", 0)
        self.image_external_id = node.create_state.get(
            "image_external_id", image_external_id)
        if self.image_external_id:
            self.image_reference = {'id': self.image_external_id}
        else:
            self.image_reference = None
        self.subnet_id = None

        # Restore local objects
        self.node.quota = self.endpoint.getQuotaForLabel(
            self.label, self.flavor)

        if self.state == self.VM_CREATING_SUBMIT:
            for instance in self.endpoint.listInstances():
                if instance.metadata.get('zuul_node_uuid') == node.uuid:
                    self.node.azure_vm_id = instance.azure_vm_id
            if self.node.azure_vm_id:
                self.state = self.VM_CREATING

        self.vm = None
        if self.node.azure_vm_id:
            vm = dict(
                type='Microsoft.Compute/virtualMachines',
                id=node.azure_vm_id,
            )
            self.vm = self.endpoint._refresh(vm)

    def toDict(self):
        data = super().toDict()
        data.update(
            attempts=self.attempts,
            image_external_id=self.image_external_id,
        )
        return data

    def advance(self):
        if self.state == self.START:
            # Find an appropriate image if filters were provided
            if self.image_reference is None:
                self.image_reference =\
                    self.endpoint._getImageReferenceForImage(self.image)
            if self.subnet_id is None:
                self.subnet_id =\
                    self.endpoint._getSubnetIdForLabel(
                        self.resource_group, self.label)
            self.state = self.VM_CREATING_SUBMIT

        if self.state == self.VM_CREATING_SUBMIT:
            if not self.create_future:
                self.create_future = self.endpoint._submitApi(
                    self.endpoint._createVirtualMachine,
                    self.resource_group, self.label, self.flavor,
                    self.image, self.image_reference, self.subnet_id,
                    self.tags, self.hostname, self.log)

            self.vm = self.endpoint._completeApi(self.create_future)
            if self.vm is None:
                return
            self.node.azure_vm_id = self.vm['id']
            self.state = self.VM_CREATING

        if self.state == self.VM_CREATING:
            self.vm = self.endpoint._refresh(self.vm)
            if self.endpoint._succeeded(self.vm):
                self.nic = (self.vm['properties']['networkProfile']
                            ['networkInterfaces'][0]).copy()
                self.nic['type'] = 'Microsoft.Network/networkInterfaces'
                self.state = self.NIC_QUERY
            elif self.endpoint._failed(self.vm):
                raise exceptions.LaunchStatusException("VM in failed state")
            else:
                return

        if self.state == self.NIC_QUERY:
            self.nic = self.endpoint._refresh(self.nic, force=True)
            all_found = True
            for ip_config_data in self.nic['properties']['ipConfigurations']:
                ip_config_prop = ip_config_data['properties']
                if 'privateIPAddress' not in ip_config_prop:
                    all_found = False
            if all_found:
                for ip_config in self.nic['properties']['ipConfigurations']:
                    if 'publicIPAddress' in ip_config['properties']:
                        public_ip = ip_config['properties']['publicIPAddress']
                        ip_version = (ip_config['properties']
                                      ['privateIPAddressVersion'])
                        if ip_version == 'IPv4':
                            self.public_ipv4 = public_ip
                            self.public_ipv4['type'] =\
                                'Microsoft.Network/publicIPAddresses'
                        if ip_version == 'IPv6':
                            self.public_ipv6 = public_ip
                            self.public_ipv6['type'] =\
                                'Microsoft.Network/publicIPAddresses'
                self.state = self.PIP_QUERY

        if self.state == self.PIP_QUERY:
            all_found = True
            if self.public_ipv4:
                self.public_ipv4 = self.endpoint._refresh(
                    self.public_ipv4, force=True)
                if 'ipAddress' not in self.public_ipv4['properties']:
                    all_found = False
            if self.public_ipv6:
                self.public_ipv6 = self.endpoint._refresh(
                    self.public_ipv6, force=True)
                if 'ipAddress' not in self.public_ipv6['properties']:
                    all_found = False
            if all_found:
                self.state = self.COMPLETE

        if self.state == self.COMPLETE:
            self.complete = True
            return AzureInstance(self.endpoint.region, self.vm,
                                 self.node.quota, self.nic,
                                 self.public_ipv4, self.public_ipv6)


class AzureImageUploadJob(BaseImageUploadJob):
    def __init__(self, endpoint, resource_group,
                 provider_image, image_name,
                 image_format, metadata,
                 timeout):
        super().__init__()
        self.endpoint = endpoint
        self.resource_group = resource_group
        self.provider_image = provider_image
        self.image_name = image_name
        self.image_format = image_format
        self.metadata = metadata
        self.timeout = timeout

    def run(self, filename):
        self.endpoint.log.debug("Uploading image %s",
                                self.image_name)

        image_id = self.endpoint._uploadImageSnapshot(
            self.resource_group, self.provider_image, self.image_name,
            filename, self.image_format, self.metadata, self.timeout)
        return image_id


class AzureSnapshotUploader(ImageUploader):
    segment_size = 4 * MIB

    def __init__(self, resource_group, *args, **kw):
        super().__init__(*args, **kw)
        self.resource_group = resource_group

    def uploadSegment(self, segment):
        data = segment.data
        start = segment.offset
        end = start + len(data) - 1
        self.retry(
            self.endpoint._client.upload_sas_chunk,
            self.url, start, end, data
        )

    def startUpload(self):
        disk_info = {
            "location": self.endpoint.region,
            "tags": self.metadata,
            "properties": {
                "creationData": {
                    "createOption": "Upload",
                    "uploadSizeBytes": self.size,
                }
            }
        }
        self.log.debug("Creating disk for image upload")

        with self.endpoint.rate_limiter:
            r = self.endpoint._client.resource_group_disks.create(
                self.resource_group, self.image_name, disk_info)
        r = self.endpoint._client.wait_for_async_operation(r)

        if r['status'] != 'Succeeded':
            raise Exception("Unable to create disk for image upload")
        self.disk_id = r['properties']['output']['id']

        disk_grant = {
            "access": "Write",
            "durationInSeconds": 24 * 60 * 60,
        }
        self.log.debug("Enabling write access to disk for image upload")
        with self.endpoint.rate_limiter:
            r = self.endpoint._client.resource_group_disks.post(
                self.resource_group, self.image_name,
                'beginGetAccess', disk_grant)
        r = self.endpoint._client.wait_for_async_operation(r)

        if r['status'] != 'Succeeded':
            raise Exception("Unable to begin write access on disk")
        self.url = r['properties']['output']['accessSAS']
        self.log.debug("Uploading image")

    def finishUpload(self):
        disk_grant = {}
        self.log.debug("Disabling write access to disk for image upload")
        with self.endpoint.rate_limiter:
            r = self.endpoint._client.resource_group_disks.post(
                self.resource_group, self.image_name,
                'endGetAccess', disk_grant)
        r = self.endpoint._client.wait_for_async_operation(r)

        if r['status'] != 'Succeeded':
            raise Exception("Unable to end write access on disk")

        image_info = {
            "location": self.endpoint.region,
            "tags": self.metadata,
            "properties": {
                "hyperVGeneration": "V2",
                "storageProfile": {
                    "osDisk": {
                        "osType": "Linux",
                        "managedDisk": {
                            "id": self.disk_id,
                        },
                        "osState": "Generalized"
                    },
                    "zoneResilient": True
                }
            }
        }
        self.log.debug("Creating image from disk")
        with self.endpoint.rate_limiter:
            image = self.endpoint._client.resource_group_images.create(
                self.resource_group, self.image_name, image_info)
        image = self.endpoint._client.wait_for_async_operation(image)

        if image['status'] != 'Succeeded':
            raise Exception("Unable to create image from disk")

        self.log.debug("Deleting disk for image upload")
        with self.endpoint.rate_limiter:
            r = self.endpoint._client.resource_group_disks.delete(
                self.resource_group,
                self.image_name)
        r = self.endpoint._client.wait_for_async_operation(r)

        if r['status'] != 'Succeeded':
            raise Exception("Unable to delete disk for image upload")

        return image['properties']['output']['id']

    def abortUpload(self):
        try:
            self.finishUpload()
        except Exception:
            pass
        with self.endpoint.rate_limiter:
            self.endpoint._client.resource_group_disks.delete(
                self.resource_group,
                self.image_name)


class AzureProviderEndpoint(BaseProviderEndpoint):
    """An Azure Endpoint corresponds to a single Azure region, and can include
    multiple availability zones."""

    IMAGE_UPLOAD_SLEEP = 30

    def __init__(self, zk_client, driver, connection, region, system_id):
        name = f'{connection.connection_name}-{region}'
        super().__init__(zk_client, driver, connection, name, system_id)
        self.log = logging.getLogger(f"zuul.azure.{self.name}")
        self.region = region

        self.rate_limiter = RateLimiter(self.name,
                                        connection.rate)

        self.image_id_by_filter_cache = cachetools.TTLCache(
            maxsize=8192, ttl=(5 * 60))

    def _getClient(self):
        return azul.AzureCloud(
            subscription_id=self.connection.subscription_id,
            tenant_id=self.connection.tenant_id,
            client_id=self.connection.client_id,
            client_secret=self.connection.client_secret,
            federated_token_file=self.connection.federated_token_file,
        )

    def startEndpoint(self):
        self._running = True
        self.log.debug("Starting Azure endpoint")
        self._client = self._getClient()

        # The default http connection pool size is 10; match it for
        # efficiency.
        workers = 10
        self.log.info("Create executor with max workers=%s", workers)
        self.api_executor = ThreadPoolExecutor(
            thread_name_prefix=f'azure-api-{self.name}',
            max_workers=workers)

        # Use a lazy TTL cache for these.  This uses the TPE to
        # asynchronously update the cached values, meanwhile returning
        # the previous cached data if available.  This means every
        # call after the first one is instantaneous.
        self._listPublicIPAddresses = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listPublicIPAddresses)
        self._listNetworkInterfaces = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listNetworkInterfaces)
        self._listVirtualMachines = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listVirtualMachines)
        self._listDisks = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listDisks)
        self._listImages = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listImages)

        self.skus = {}
        self._getSKUs()

    def stopEndpoint(self):
        self.log.debug("Stopping Azure endpoint")
        self.api_executor.shutdown()
        self._running = False

    def listResources(self, providers):
        resource_groups = set()
        for provider in providers:
            resource_groups.add(provider.resource_group)

        for vm in self._listVirtualMachines():
            yield AzureResource(vm.get('tags', {}),
                                AzureResource.TYPE_INSTANCE,
                                vm['id'])
        for nic in self._listNetworkInterfaces():
            yield AzureResource(nic.get('tags', {}),
                                AzureResource.TYPE_NIC, nic['id'])
        for pip in self._listPublicIPAddresses():
            yield AzureResource(pip.get('tags', {}),
                                AzureResource.TYPE_PIP, pip['id'])
        for disk in self._listDisks():
            yield AzureResource(disk.get('tags', {}),
                                AzureResource.TYPE_DISK, disk['id'])
        for image in self._listImages():
            yield AzureResource(image.get('tags', {}),
                                AzureResource.TYPE_IMAGE, image['id'])

    def deleteResource(self, resource):
        self.log.info("Deleting leaked %s: %s",
                      resource.type, resource.id)
        if resource.type == AzureResource.TYPE_INSTANCE:
            crud = self._client.virtual_machines
        elif resource.type == AzureResource.TYPE_NIC:
            crud = self._client.network_interfaces
        elif resource.type == AzureResource.TYPE_PIP:
            crud = self._client.public_ip_addresses
        elif resource.type == AzureResource.TYPE_DISK:
            crud = self._client.disks
        elif resource.type == AzureResource.TYPE_IMAGE:
            crud = self._client.images
        with self.rate_limiter:
            try:
                crud.delete_by_id(resource.id)
            except azul.AzureNotFoundError:
                pass

    def listInstances(self):
        for vm in self._listVirtualMachines():
            sku = self.skus.get((vm['properties']['hardwareProfile']['vmSize'],
                                 vm['location']))
            spot = vm.get('properties', {}).get('priority', '') == 'Spot'
            quota = quota_info_from_sku(sku, spot)
            yield AzureInstance(self.region, vm, quota)

    def getQuotaForLabel(self, label, flavor):
        # This should not rely on the client connection, only the
        # quota cache.
        check_vm_size = flavor.vm_size
        key = f'vm-{flavor.priority}-{check_vm_size}'
        quota = self.quota_cache.getResource(key)
        return quota

    def refreshQuotaLimits(self, update):
        if self.quota_cache.hasLimits() and not update:
            return False

        with self.rate_limiter:
            r = self._client.compute_usages.list(self.region)
        cores = spot_cores = instances = math.inf
        for item in r:
            if item['name']['value'] == 'cores':
                cores = item['limit']
            if item['name']['value'] == 'lowPriorityCores':
                spot_cores = item['limit']
            elif item['name']['value'] == 'virtualMachines':
                instances = item['limit']
        limits = QuotaInformation(cores=cores,
                                  spot_cores=spot_cores,
                                  instances=instances,
                                  default=math.inf)
        self.quota_cache.setLimits(limits)
        return True

    def refreshQuotaForLabel(self, label, flavor, update):
        key = f'vm-{flavor.priority}-{flavor.vm_size}'
        if update or not self.quota_cache.hasResource(key):
            sku = self.skus.get((flavor.vm_size, self.region))
            spot = flavor.priority == 'spot'
            quota = quota_info_from_sku(sku, spot)
            self.quota_cache.setResource(key, quota)

    def getImageUploadJob(self, resource_group, provider_image,
                          image_name, image_format, metadata, md5,
                          sha256):
        timeout = provider_image.import_timeout
        return AzureImageUploadJob(
            self, resource_group,
            provider_image, image_name,
            image_format, metadata,
            timeout)

    def _uploadImageSnapshot(self, resource_group, provider_image,
                             image_name, filename, image_format,
                             metadata, timeout):
        # Import snapshot
        uploader = AzureSnapshotUploader(
            resource_group, self, self.log, filename,
            image_name, metadata)
        self.log.debug("Uploading image %s", image_name)
        image_id = uploader.upload(timeout)

        self.log.debug("Upload of image %s complete as %s",
                       image_name, image_id)
        return image_id

    # Local implementation below

    def _submitApi(self, api, *args, **kw):
        return self.api_executor.submit(
            api, *args, **kw)

    def _completeApi(self, future):
        if not future.done():
            return None
        return future.result()

    @staticmethod
    def _succeeded(obj):
        return obj.get('properties', {}).get(
            'provisioningState') == 'Succeeded'

    @staticmethod
    def _failed(obj):
        return obj.get('properties', {}).get(
            'provisioningState') == 'Failed'

    def _refresh(self, obj, force=False):
        if self._succeeded(obj) and not force:
            return obj

        if obj['type'] == 'Microsoft.Network/publicIPAddresses':
            l = self._listPublicIPAddresses()
        if obj['type'] == 'Microsoft.Network/networkInterfaces':
            l = self._listNetworkInterfaces()
        if obj['type'] == 'Microsoft.Compute/virtualMachines':
            l = self._listVirtualMachines()

        for new_obj in l:
            if new_obj['id'] == obj['id']:
                return new_obj
        return obj

    def _refreshDelete(self, obj):
        if obj is None:
            return obj

        if obj['type'] == 'Microsoft.Network/publicIPAddresses':
            l = self._listPublicIPAddresses()
        if obj['type'] == 'Microsoft.Network/networkInterfaces':
            l = self._listNetworkInterfaces()
        if obj['type'] == 'Microsoft.Compute/virtualMachines':
            l = self._listVirtualMachines()

        for new_obj in l:
            if new_obj['id'] == obj['id']:
                return new_obj
        return None

    def _getSKUs(self):
        self.log.debug("Querying compute SKUs")
        with self.rate_limiter:
            for sku in self._client.compute_skus.list():
                for location in sku['locations']:
                    key = (sku['name'], location)
                    self.skus[key] = sku
        self.log.debug("Done querying compute SKUs")

    def _listPublicIPAddresses(self):
        with self.rate_limiter:
            return self._client.public_ip_addresses.list()

    def _listNetworkInterfaces(self):
        with self.rate_limiter:
            return self._client.network_interfaces.list()

    def _listVirtualMachines(self):
        with self.rate_limiter:
            return self._client.virtual_machines.list()

    def _createVirtualMachine(self, resource_group, label, flavor,
                              image, image_reference, subnet_id, tags,
                              hostname, log):
        os_profile = {'computerName': hostname}
        if label.key_data:
            linux_config = {
                'ssh': {
                    'publicKeys': [{
                        'path': "/home/%s/.ssh/authorized_keys" % (
                            image.username),
                        'keyData': label.key_data,
                    }]
                },
                "disablePasswordAuthentication": True,
            }
            os_profile['linuxConfiguration'] = linux_config
        if image.username:
            os_profile['adminUsername'] = image.username
        if image.generate_password:
            os_profile['adminPassword'] = generate_password()

        # Nodepool provided a plaintext password option which doesn't
        # make sense in NIZ.  It also supported the "customData"
        # attribute but Azure recommends "userData" instead, and that
        # matches other clouds, so we only support that.

        # Begin disk config

        os_disk = {
            'createOption': 'FromImage',
            'deleteOption': 'Delete',
        }
        if label.volume_size:
            os_disk['diskSizeGB'] = label.volume_size

        if label.ephemeral_disk:
            # Caching must be set to ReadOnly for ephemeral disks
            # https://learn.microsoft.com/en-us/azure/virtual-machines/ephemeral-os-disks-deploy
            os_disk['caching'] = 'ReadOnly'
            os_disk['diffDiskSettings'] = {
                # This causes the disk to be ephemeral
                'option': 'Local',
                # Another frequently used option is "placement", but
                # if it is omitted, Azure will pick the correct one
                # for the VM based on the SKU.
                # https://learn.microsoft.com/en-us/azure/virtual-machines/ephemeral-os-disks#placement-options-for-ephemeral-os-disks
            }

        # Begin network config
        # There are two parameters for IP addresses: SKU and
        # allocation method.  SKU is "basic" or "standard".
        # Allocation method is "static" or "dynamic".  Between IPv4
        # and v6, SKUs cannot be mixed (the same sku must be used for
        # both protocols).  The standard SKU only supports static
        # allocation.  Static is cheaper than dynamic, but basic is
        # cheaper than standard.  Also, dynamic is faster than static.
        # Basic is being retired, and static allocation (contrary to
        # what it sounds like) works for our use case.
        # Therefore, we use Standird + static for everything.

        def make_ip_config(name, version, subnet_id, pip):
            ip_config = {
                'name': f'{name}-config',
                'properties': {
                    'privateIPAddressVersion': version,
                    'subnet': {
                        'id': subnet_id
                    },
                }
            }
            if pip:
                ip_config['properties']['publicIPAddressConfiguration'] = {
                    'name': f'{name}-ip',
                    'tags': tags,
                    'properties': {
                        'deleteOption': 'Delete',
                        'publicIPAllocationMethod': 'Static',
                        'publicIPAddressVersion': version,
                    },
                    'sku': {
                        'name': 'Standard',
                    },
                }
            return ip_config

        ip_configs = []

        ipv6 = flavor.ipv6 or flavor.public_ipv6
        ipv4 = flavor.ipv4 or flavor.public_ipv4 or not ipv6

        if ipv4:
            ip_configs.append(make_ip_config(f'{hostname}-v4',
                                             'IPv4', subnet_id,
                                             flavor.public_ipv4))
        if ipv6:
            ip_configs.append(make_ip_config(f'{hostname}-v6',
                                             'IPv6', subnet_id,
                                             flavor.public_ipv6))

        nic_config = {
            'name': f'{hostname}-nic',
            'tags': tags,
            'properties': {
                'deleteOption': 'Delete',
                'ipConfigurations': ip_configs
            }
        }

        spec = {
            'location': self.region,
            'tags': tags,
            'properties': {
                'osProfile': os_profile,
                'hardwareProfile': {
                    'vmSize': flavor.vm_size,
                },
                'storageProfile': {
                    'imageReference': image_reference,
                    'osDisk': os_disk,
                },
                'networkProfile': {
                    'networkApiVersion': (
                        self._client.network_interfaces.args['apiVersion']),
                    'networkInterfaceConfigurations': [nic_config],
                },
            },
        }
        if label.userdata:
            spec['properties']['userData'] = label.userdata

        # build resource id for all configured User-assigned Identities
        uai_resource_ids = set()
        for uai in label.user_assigned_identities:
            uai_rg_name = uai.get("resource-group", resource_group)
            uai_id = self._client.managed_identities.id(
                resourceGroupName=uai_rg_name, resourceName=uai['name'])
            uai_resource_ids.add(uai_id)

        # adding empty userAssignedIdentities is not allowed by Azure
        if uai_resource_ids:
            spec['identity'] = {
                'type': 'UserAssigned',
                'userAssignedIdentities': {
                    rid: {} for rid in uai_resource_ids
                },
            }

        if flavor.priority == 'spot':
            spec['properties']['evictionPolicy'] = 'Delete'
            spec['properties']['priority'] = 'Spot'

        with self.rate_limiter:
            log.debug(f"Creating VM {hostname}")
            return self._client.resource_group_virtual_machines.create(
                resource_group, hostname, spec)

    def _deleteVirtualMachine(self, vm_id):
        for vm in self._listVirtualMachines():
            if vm['id'] == vm_id:
                break
        else:
            self.log.warning("VM not found when deleting %s", vm_id)
            return None
        with self.rate_limiter:
            self.log.debug("Deleting VM %s", vm_id)
            self._client.virtual_machines.delete_by_id(vm_id)
        return vm

    def _listDisks(self):
        with self.rate_limiter:
            return self._client.disks.list()

    def _listImages(self):
        with self.rate_limiter:
            return self._client.images.list()

    def _getImageReferenceFromFilter(self, image_filter):
        # Normally we would decorate this method, but our cache key is
        # complex, so we serialize it to JSON and manage the cache
        # ourselves.
        cache_key = json.dumps(image_filter)
        if val := self.image_id_by_filter_cache.get(cache_key):
            return val

        images = self._listImages()
        images = [i for i in images
                  if i['properties']['provisioningState'] == 'Succeeded']
        if image_filter['name']:
            images = [i for i in images
                      if i['name'] == image_filter['name']]
        if image_filter['location']:
            images = [i for i in images
                      if i['location'] == image_filter['location']]
        if image_filter['tags']:
            for k, v in image_filter['tags'].items():
                images = [i for i in images if i['tags'].get(k) == v]
        images = sorted(images, key=lambda i: i['name'])
        if not images:
            raise Exception("Unable to find image matching filter: %s",
                            image_filter)
        image = images[-1]
        self.log.debug("Found image matching filter: %s", image)
        return {'id': image['id']}

    def _getImageReferenceForImage(self, image):
        if image.image_id:
            return {'id': image.image_id}
        if image.image_reference:
            return image.image_reference
        if image.image_filter:
            return self._getImageReferenceFromFilter(image.image_filter)
        if g := image.community_gallery_image:
            gallery_image_id = (
                f"/CommunityGalleries/{g['gallery_name']}"
                f"/Images/{g['name']}"
            )
            if 'version' in g:
                gallery_image_id += f"/Versions/{g['version']}"
            image_reference = {
                'communityGalleryImageId': gallery_image_id
            }
            return image_reference
        if g := image.shared_gallery_image:
            gallery_image_id = (
                f"/SharedGalleries/{g['gallery_name']}"
                f"/Images/{g['name']}"
            )
            if 'version' in g:
                gallery_image_id += f"/Versions/{g['version']}"
            image_reference = {
                'sharedGalleryImageId': gallery_image_id
            }
            return image_reference

    def _getSubnetIdForLabel(self, resource_group, label):
        if label.subnet_id:
            return label.subnet_id

        ref = label.subnet_reference
        subnet_id = self._client.subnets.id(
            resourceGroupName=ref.get('resource-group', resource_group),
            virtualNetworkName=ref['network'],
            resourceName=ref['subnet'],
        )
        return subnet_id



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import voluptuous as vs


FILE_COMMENT = {
    'line': int,
    'message': str,
    'range': {
        'start_line': int,
        'start_character': int,
        'end_line': int,
        'end_character': int,
    },
    'level': str,
}

FILE_COMMENTS = {str: [FILE_COMMENT]}

FILE_COMMENTS_SCHEMA = vs.Schema(FILE_COMMENTS)


def validate(file_comments):
    FILE_COMMENTS_SCHEMA(file_comments)


def extractLines(file_comments):
    """Extract file/line tuples from file comments for mapping"""

    lines = set()
    for path, comments in file_comments.items():
        for comment in comments:
            if 'line' in comment:
                lines.add((path, int(comment['line'])))
            if 'range' in comment:
                rng = comment['range']
                for key in ['start_line', 'end_line']:
                    if key in rng:
                        lines.add((path, int(rng[key])))
    return list(lines)


def updateLines(file_comments, lines):
    """Update the line numbers in file_comments with the supplied mapping"""

    for path, comments in file_comments.items():
        for comment in comments:
            if 'line' in comment:
                comment['line'] = lines.get((path, comment['line']),
                                            comment['line'])
            if 'range' in comment:
                rng = comment['range']
                for key in ['start_line', 'end_line']:
                    if key in rng:
                        rng[key] = lines.get((path, rng[key]), rng[key])



# Copyright 2019 OpenStack Foundation
# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import math
import threading
import time
import jwt
import requests
from urllib.parse import urljoin

from zuul import exceptions
from zuul.driver import AuthenticatorInterface
from zuul.lib.config import any_to_bool


logger = logging.getLogger("zuul.auth.jwt")

# A few notes on differences between the OpenID Connect specification
# and OIDC as implemented by Microsoft, which necessitates several
# extra configuration options:
#
# 1) The issuer (iss) returned in the JWT is not the authority, but
# rather a URL referring to the AD instance, therefore authority must
# be specified separately.
# 2) The audience (aud) returned in the JWT is not the client_id, but
# rather a URL constructed from the client_id, therefore must also be
# specified separately.
# 3) By default, the JWT is simply a forwarded copy of a token from
# the Microsoft Graph service.  It is signed by the graph service and
# not the authority as requested, it therefore fails signature
# validation.  In order to cause the authority to generate its own
# token signed by the expected keys, we must create a new scope
# (api://.../zuul) and request that scope with the token.
# 4) The userinfo service referred to by the JWT is the Microsoft
# Graph service, which means that once our token is signed by the
# Microsoft login keys (see item #3) the graph service is then unable
# to validate the token.  Therefore we must configure the javascript
# oidc library not to request userinfo and rely only on what is
# supplied in the token.


class JWTAuthenticator(AuthenticatorInterface):
    """The base class for JWT-based authentication."""

    def __init__(self, **conf):
        # Common configuration for all authenticators
        self.uid_claim = conf.get('uid_claim', 'sub')
        self.issuer_id = conf.get('issuer_id')
        self.authority = conf.get('authority', self.issuer_id)
        self.client_id = conf.get('client_id')
        self.audience = conf.get('audience', self.client_id)
        self.realm = conf.get('realm')
        self.allow_authz_override = any_to_bool(
            conf.get('allow_authz_override', False))
        try:
            self.skew = int(conf.get('skew', 0))
        except Exception:
            raise ValueError(
                'skew must be an integer, got %s' % conf.get('skew'))
        try:
            self.max_validity_time = float(conf.get('max_validity_time',
                                                    math.inf))
        except ValueError:
            raise ValueError('"max_validity_time" must be a numerical value')

    def get_capabilities(self):
        return {
            self.realm: {
                'authority': self.authority,
                'client_id': self.client_id,
                'type': 'JWT',
                'driver': getattr(self, 'name', 'N/A'),
            }
        }

    def _decode(self, rawToken):
        raise NotImplementedError

    def decodeToken(self, rawToken):
        """Verify the raw token and return the decoded dictionary of claims"""
        try:
            decoded = self._decode(rawToken)
        except jwt.exceptions.InvalidSignatureError:
            raise exceptions.AuthTokenInvalidSignatureException(
                realm=self.realm)
        except jwt.exceptions.DecodeError:
            raise exceptions.AuthTokenUndecodedException(
                realm=self.realm)
        except jwt.exceptions.ExpiredSignatureError:
            raise exceptions.TokenExpiredError(
                realm=self.realm)
        except jwt.exceptions.InvalidIssuerError:
            raise exceptions.IssuerUnknownError(
                realm=self.realm)
        except jwt.exceptions.InvalidAudienceError:
            raise exceptions.IncorrectAudienceError(
                realm=self.realm)
        except Exception as e:
            raise exceptions.AuthTokenUnauthorizedException(
                realm=self.realm,
                msg=e)
        # Missing claim tests
        if not all(x in decoded for x in ['aud', 'iss', 'exp', 'sub']):
            raise exceptions.MissingClaimError(realm=self.realm)
        if self.max_validity_time < math.inf and 'iat' not in decoded:
            raise exceptions.MissingClaimError(
                msg='Missing "iat" claim',
                realm=self.realm)
        if self.uid_claim not in decoded:
            raise exceptions.MissingUIDClaimError(realm=self.realm)
        # Time related tests
        expires = decoded.get('exp', 0)
        issued_at = decoded.get('iat', 0)
        now = time.time()
        if issued_at + self.skew > now:
            raise exceptions.AuthTokenUnauthorizedException(
                msg='"iat" claim set in the future',
                realm=self.realm
            )
        if now - issued_at > self.max_validity_time:
            raise exceptions.TokenExpiredError(
                msg='Token was issued too long ago',
                realm=self.realm)
        if expires + self.skew < now:
            raise exceptions.TokenExpiredError(realm=self.realm)
        # Zuul-specific claims tests
        zuul_claims = decoded.get('zuul', {})
        admin_tenants = zuul_claims.get('admin', [])
        if not isinstance(admin_tenants, list):
            raise exceptions.IncorrectZuulAdminClaimError(realm=self.realm)
        if admin_tenants and not self.allow_authz_override:
            msg = ('Issuer "%s" attempt to override User "%s" '
                   'authorization denied')
            logger.info(msg % (decoded['iss'], decoded[self.uid_claim]))
            logger.debug('%r' % admin_tenants)
            raise exceptions.UnauthorizedZuulAdminClaimError(
                realm=self.realm)
        if admin_tenants and self.allow_authz_override:
            msg = ('Issuer "%s" attempt to override User "%s" '
                   'authorization granted')
            logger.info(msg % (decoded['iss'], decoded[self.uid_claim]))
            logger.debug('%r' % admin_tenants)
        return decoded

    def authenticate(self, rawToken):
        decoded = self.decodeToken(rawToken)
        # inject the special authenticator-specific uid
        decoded['__zuul_uid_claim'] = decoded[self.uid_claim]
        return decoded


class HS256Authenticator(JWTAuthenticator):
    """JWT authentication using the HS256 algorithm.

    Requires a shared secret between Zuul and the identity provider."""

    name = algorithm = 'HS256'

    def __init__(self, **conf):
        super(HS256Authenticator, self).__init__(**conf)
        self.secret = conf.get('secret')

    def _decode(self, rawToken):
        return jwt.decode(rawToken, self.secret, issuer=self.issuer_id,
                          audience=self.audience,
                          algorithms=[self.algorithm])


class RS256Authenticator(JWTAuthenticator):
    """JWT authentication using the RS256 algorithm.

    Requires a copy of the public key of the identity provider."""

    name = algorithm = 'RS256'

    def __init__(self, **conf):
        super(RS256Authenticator, self).__init__(**conf)
        with open(conf.get('public_key')) as pk:
            self.public_key = pk.read()

    def _decode(self, rawToken):
        return jwt.decode(rawToken, self.public_key, issuer=self.issuer_id,
                          audience=self.audience,
                          algorithms=[self.algorithm])


class OpenIDConnectAuthenticator(JWTAuthenticator):
    """JWT authentication using an OpenIDConnect provider.

    If the optional 'keys_url' parameter is not specified, the authenticator
    will attempt to determine it via the well-known configuration URI as
    described in
    https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderConfig"""  # noqa

    # default algorithm, TODO: should this be a config param?
    algorithm = 'RS256'
    name = 'OpenIDConnect'

    def __init__(self, **conf):
        super(OpenIDConnectAuthenticator, self).__init__(**conf)
        self.keys_url = conf.get('keys_url', None)
        self.scope = conf.get('scope', 'openid profile')
        self.load_user_info = any_to_bool(
            conf.get('load_user_info', True))
        self._init_lock = threading.Lock()
        self._client = None

    def get_client(self):
        if self._client is not None:
            return self._client
        with self._init_lock:
            if self._client is not None:
                return self._client
            keys_url = self.keys_url
            if keys_url is None:
                well_known = self.get_well_known_config()
                keys_url = well_known.get('jwks_uri', None)
            if keys_url is None:
                msg = 'Invalid OpenID configuration: "jwks_uri" not found'
                logger.error(msg)
                raise exceptions.JWKSException(
                    realm=self.realm,
                    msg=msg)
            self._client = jwt.PyJWKClient(keys_url)
            return self._client

    def get_key(self, key_id):
        # This has its own exception handler
        client = self.get_client()
        try:
            signing_key = client.get_signing_key(kid=key_id)
        except Exception as e:
            msg = 'Could not fetch Identity Provider keys at %s: %s'
            logger.error(msg % (client.uri, e))
            raise exceptions.JWKSException(
                realm=self.realm,
                msg='There was an error while fetching '
                    'keys for Identity Provider, check logs for details')
        algorithm = signing_key._jwk_data.get("alg", None) or self.algorithm
        key = signing_key.key
        return key, algorithm

    def get_well_known_config(self):
        issuer = self.issuer_id
        if not issuer.endswith('/'):
            issuer += '/'
        well_known_uri = urljoin(issuer,
                                 '.well-known/openid-configuration')
        try:
            return requests.get(well_known_uri).json()
        except Exception as e:
            msg = 'Could not fetch OpenID configuration at %s: %s'
            logger.error(msg % (well_known_uri, e))
            raise exceptions.JWKSException(
                realm=self.realm,
                msg='There was an error while fetching '
                    'OpenID configuration, check logs for details')

    def get_capabilities(self):
        d = super(OpenIDConnectAuthenticator, self).get_capabilities()
        d[self.realm]['scope'] = self.scope
        d[self.realm]['load_user_info'] = self.load_user_info
        return d

    def _decode(self, rawToken):
        unverified_headers = jwt.get_unverified_header(rawToken)
        key_id = unverified_headers.get('kid', None)
        if key_id is None:
            raise exceptions.JWKSException(
                self.realm, 'No key ID in token header')
        key, algorithm = self.get_key(key_id)
        return jwt.decode(rawToken, key, issuer=self.issuer_id,
                          audience=self.audience,
                          algorithms=[algorithm])


AUTHENTICATORS = {
    'HS256': HS256Authenticator,
    'RS256': RS256Authenticator,
    'RS256withJWKS': OpenIDConnectAuthenticator,
    'OpenIDConnect': OpenIDConnectAuthenticator,
}


def get_authenticator_by_name(name):
    if name == 'RS256withJWKS':
        logger.info(
            'Driver "%s" is deprecated, please use "OpenIDConnect" instead')
    return AUTHENTICATORS[name]
